import configparser
import shutil
import sys

from pathlib import Path
from typing import Dict, Optional

from semver import Version

from qgis_plugin_manager import echo
from qgis_plugin_manager.definitions import Plugin
from qgis_plugin_manager.utils import (
    PluginManagerError,
    get_default_remote_repository,
    get_semver_version,
    similar_names,
    sources_file,
)


class LocalDirectory:
    def __init__(self, folder: Path):
        """Constructor"""
        self.folder = folder
        # Dictionary : folder : plugin name
        self._plugins: Dict[str, str] = {}
        self._plugins_metadata: Dict[str, configparser.SectionProxy] = {}
        self.list_plugins()

    def init(self, qgis_version: Optional[str]) -> bool:
        """Init this qgis-plugin-manager by creating the default sources.list."""
        source_file = sources_file(self.folder)
        if source_file.exists():
            echo.alert(f"{source_file.absolute()} is already existing. Quit")
            return False

        if qgis_version:
            try:
                version: Optional[str] = str(get_semver_version(qgis_version))
            except Exception:
                raise PluginManagerError(f"{qgis_version} is not a valid QGIS version") from None
        else:
            version = "[VERSION]"

        repository = get_default_remote_repository()

        print(f"Init {repository}", file=sys.stderr)

        server = f"https://plugins.qgis.org/plugins/plugins.xml?qgis={version}\n"

        try:
            with open(source_file, "w", encoding="utf8") as f:
                f.write(server)
        except PermissionError:
            # https://github.com/3liz/qgis-plugin-manager/issues/53
            echo.critical("The directory is not writable.")
            return False

        echo.info(f"{source_file.absolute()} has been written.")
        return True

    def plugin_list(self) -> Dict[str, str]:
        return self._plugins

    def list_plugins(self):
        """Get the list of plugins installed in the current directory."""
        for folder in self.folder.iterdir():
            if not folder.is_dir():
                continue

            # Skip hidden folders
            if folder.name.startswith("."):
                continue

            have_python = folder.joinpath("__init__.py").exists()
            have_metadata = folder.joinpath("metadata.txt").exists()
            if have_python and have_metadata:
                try:
                    metadata = self._get_plugin_metadata(folder.name)
                    self._plugins[folder.name] = metadata["name"]
                    self._plugins_metadata[folder.name] = metadata
                except KeyError:
                    echo.alert(f"WARNING: invalid metadata found in {folder}")

    def _get_plugin_metadata(self, plugin_folder: str) -> configparser.SectionProxy:
        """For a given plugin installed, get a metadata item."""
        config_parser = configparser.ConfigParser()
        with self.folder.joinpath(f"{plugin_folder}", "metadata.txt").open(encoding="utf8") as f:
            config_parser.read_file(f)
            return config_parser["general"]

    def get_plugin_folder_from_name(self, plugin_name: str) -> Optional[str]:
        for folder, name in self._plugins.items():
            if plugin_name == name:
                return folder
        return None

    def plugin_info(self, plugin: str) -> Optional[Plugin]:
        """For a given plugin, retrieve all metadata."""
        if plugin in self._plugins:
            # It's plugin folder
            plugin_folder: Optional[str] = plugin
        else:
            plugin_folder = self.get_plugin_folder_from_name(plugin)

        # Type narrowing
        if not plugin_folder:
            # No plugin
            return None

        md = self._plugins_metadata[plugin_folder]

        def maybe_version(ver: Optional[str]) -> Optional[Version]:
            return get_semver_version(ver) if ver else None

        # Make sure that version is semver compatible
        qgis_minimum_version = maybe_version(md.get("qgisMinimumVersion"))
        qgis_maximum_version = maybe_version(md.get("qgisMaximumVersion"))

        return Plugin(
            name=md["name"],
            version=get_semver_version(md.get("version") or "0.0.0."),
            experimental=md.getboolean("experimental", False),
            qgis_minimum_version=qgis_minimum_version,
            qgis_maximum_version=qgis_maximum_version,
            author_name=md.get("author"),
            server=md.getboolean("server", False),
            has_processing=md.getboolean("hasProcessingProvider", False),
            install_folder=plugin_folder,
        )

    def remove(self, plugin_name: str) -> bool:
        """Remove a plugin by its human name."""

        all_names: set[str] = set()

        for plugin_folder in self.plugin_list():
            info = self.plugin_info(plugin_folder)
            if not info:
                echo.alert(f"Cannot get plugin info for {plugin_folder}")
                continue

            # We fill all names available
            all_names.add(info.name)

            if info.name == plugin_name:
                plugin_path = self.folder.joinpath(plugin_folder)
                try:
                    shutil.rmtree(plugin_path)
                except Exception as e:
                    echo.critical(f"Plugin {plugin_name} could not be removed : {e!s}")

                if not Path(self.folder.joinpath(plugin_folder)).exists():
                    echo.success(f"Plugin {plugin_name} removed")
                    return True
                else:
                    echo.alert(
                        f"Plugin {plugin_name} using folder {plugin_folder} "
                        "could not be removed "
                        "for unknown reason."
                    )
                break
        echo.alert(f"Plugin name '{plugin_name}' not found")

        similarity = similar_names(plugin_name.lower(), list(all_names))
        for plugin in similarity:
            echo.info(f"Do you mean maybe '{plugin}' ?")

        return False



import argparse
import os

from argparse import Namespace
from pathlib import Path
from typing import (
    Callable,
    Iterator,
    Optional,
    Sequence,
)

from semver import Version

from qgis_plugin_manager import echo
from qgis_plugin_manager.definitions import Plugin
from qgis_plugin_manager.local_directory import LocalDirectory
from qgis_plugin_manager.remote import PluginNotFoundError, Remote
from qgis_plugin_manager.utils import (
    PluginManagerError,
    get_semver_version,
    install_epilog,
    print_json,
    print_table,
    qgis_server_version,
)

cli = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)


cli.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Activate verbose (debug) mode",
)


subparsers = cli.add_subparsers(
    title="commands",
    description="qgis-plugin-manager command",
)


def set_default_from_env(kwargs: dict):
    env = kwargs.pop("env", None)
    if env:
        value = os.getenv(env)
        if value is not None:
            if kwargs.get("action") == "store_true":
                kwargs["default"] = value.lower() in ("yes", "1", "y", "true", "t")
            else:
                kwargs["default"] = value


def command(name: str, **kwargs) -> Callable:
    """Wrap subcommand function"""

    def decorator(fun):
        if isinstance(fun, tuple):
            func, options = fun
        else:
            func = fun
            options = ()
        parser = subparsers.add_parser(name, description=func.__doc__, **kwargs)
        for opt in options:
            set_default_from_env(opt[1])
            parser.add_argument(*opt[0], **opt[1])
        parser.set_defaults(func=func)

    return decorator


def argument(*args, **kwargs):
    arg = (args, kwargs)

    def decorator(fun):
        if isinstance(fun, tuple):
            # tuple(Callable, tuple[option])
            return (fun[0], (*fun[1], arg))
        else:
            return (fun, (arg,))

    return decorator


#
# Context utils
#


def get_plugin_path() -> Path:
    """Get the plugin path"""
    qgis_plugin_path = os.environ.get("QGIS_PLUGINPATH")
    if qgis_plugin_path:
        # Except if the QGIS_PLUGINPATH is set
        plugin_path = Path(qgis_plugin_path)
        echo.info(f"Plugin's directory set by environment variable : {plugin_path.absolute()}\n")
    else:
        plugin_path = Path(".")
        echo.info(f"Plugin's directory set to current directory : {plugin_path.absolute()}\n")
    return plugin_path


#
# Commands
#


# Version
@command("version", help="Show version informations and exit")
def show_version(args):
    from importlib.metadata import version

    from . import __about__

    echo.echo(
        f"{__about__.__title__}\n"
        f"{__about__.__summary__}\n"
        f"Version: {version('qgis-plugin-manager')}\n"
        f"Author: {__about__.__author__}\n"
        f"Maintainer: {__about__.__maintainer__}\n"
        f"Copyright: {__about__.__copyright__}\n"
        f"Released under {__about__.__license__}\n"
        f"Url: {__about__.__uri__}\n"
    )


# Init
@command(
    "init",
    help="Create the `sources.list` with plugins.qgis.org as remote",
)
@argument("--qgis-version", help="Set the qgis version")
@argument("-u", "--update", action="store_true", help="Update index file")
def init_sources(args: Namespace):
    """If qgis-version is set to 'auto', then detect the current QGIS version"""
    # Local needed only, with QGIS version
    qgis_version = args.qgis_version
    if qgis_version == "auto":
        qgis_version = qgis_server_version()

    plugin_path = get_plugin_path()
    plugins = LocalDirectory(plugin_path)
    if plugins.init(qgis_version) and args.update:
        remote = Remote(plugin_path, qgis_version or qgis_server_version())
        if not remote.update():
            cli.exit(1)


# List
@command("list", help="List all plugins in the directory")
@argument("-o", "--outdated", action="store_true", help="List outdated plugins")
@argument(
    "--outdated-target",
    metavar="VERSION",
    help="""
        With the 'outdated' option, display the last version compatible with
        the QGIS version VERSION.
    """,
)
@argument(
    "--format",
    choices=("columns", "freeze", "list", "json"),
    default="columns",
    help="Select the output format",
)
@argument(
    "--pre",
    action="store_true",
    env="QGIS_PLUGIN_MANAGER_INCLUDE_PRERELEASE",
    help="""
        Include pre-release, development and experimental versions.
        Useful in conjonction with the 'outdated' option."
    """,
)
def list_plugins(args: Namespace):
    qgis = qgis_server_version()
    echo.info(f"QGIS version:  {qgis or 'Unknown'}")

    if not (args.outdated_target is None or args.outdated):
        echo.critical("'outdated-target' option is only usable with the '--outdated' option")
        exit(1)

    plugins = LocalDirectory(get_plugin_path())

    def infos():
        for folder, name in sorted(
            plugins.plugin_list().items(),
            key=lambda p: p[1],
        ):
            info = plugins.plugin_info(folder)
            if info:
                yield info

    if args.format == "freeze":
        if args.outdated:
            echo.critical("Youn cannot use the 'freeze' format with the 'outdated' option")
            exit(1)
        for info in infos():
            echo.echo(f"{info.name}=={info.version}")
    else:

        def install_folder(p: Plugin) -> str:
            if p.install_folder:
                return p.install_folder if p.install_folder != p.name else ""
            else:
                return ""

        if args.outdated:
            remote = Remote(plugins.folder, qgis_version=qgis)

            def outdated():
                for info in infos():
                    latest = remote.latest(
                        info.name,
                        include_prerelease=args.pre,
                        qgis_version=args.outdated_target,
                    )
                    if latest:
                        if latest.version <= info.version:
                            continue
                        latest_ver = str(latest.version)
                        latest_src = latest.source or ""
                    else:
                        latest_ver = "Removed"
                        latest_src = ""
                    yield (info, latest_ver, latest_src)

            outdated_list = tuple(outdated())
            if not outdated_list:
                echo.alert("No outdated plugins")
            else:
                if args.format == "list":
                    for info, _ in outdated_list:
                        echo.echo(info.name)
                elif args.format == "json":
                    print_json(
                        outdated_list,
                        (
                            ("name", lambda n: n[0].name),
                            ("version", lambda n: str(n[0].version)),
                            ("latest", lambda n: n[1]),
                            ("folder", lambda n: n[0].install_folder),
                            ("source", lambda n: n[2]),
                        ),
                    )
                else:
                    print_table(
                        outdated_list,
                        (
                            ("Name", lambda n: n[0].name),
                            ("Version", lambda n: str(n[0].version)),
                            ("Latest", lambda n: n[1]),
                            ("Folder", lambda n: install_folder(n[0])),
                            ("Source", lambda n: n[2]),
                        ),
                    )
        else:
            if args.format == "list":
                for info in infos():
                    echo.echo(info.name)
            elif args.format == "json":
                print_json(
                    infos(),
                    (
                        ("name", lambda p: p.name),
                        ("version", lambda p: str(p.version)),
                        ("folder", lambda p: p.install_folder),
                    ),
                )
            else:
                print_table(
                    tuple(infos()),
                    (
                        ("Name", lambda p: p.name),
                        ("Version", lambda p: str(p.version)),
                        ("Folder", install_folder),
                    ),
                )


# Install
@command("install", help="Install a plugin")
@argument("plugin_name", nargs="+", help="The plugin(s) to install")
@argument(
    "-f",
    "--force",
    action="store_true",
    help="Force (re)installation",
)
@argument(
    "-U",
    "--upgrade",
    action="store_true",
    help="Upgrade plugin to latest version",
)
@argument(
    "--fix-permissions",
    action="store_true",
    help="Set files permissions to 0644",
)
@argument(
    "--pre",
    action="store_true",
    env="QGIS_PLUGIN_MANAGER_INCLUDE_PRERELEASE",
    help=(
        "Include pre-release, development and experimental versions. By default,\ninstall only stable version"
    ),
)
@argument("--deprecated", action="store_true", help="Include deprecated versions")
def install_plugin(args: Namespace):
    """The version may be specified by appending the suffix '==version'.
    'plugin_name' might require quotes if there is space in its name.
    """
    plugin_path = get_plugin_path()

    qgis = qgis_server_version()
    echo.info("QGIS version:  {qgis or 'Unknown'}")

    remote = Remote(plugin_path, qgis_version=qgis)
    plugins = LocalDirectory(plugin_path)

    installed = 0

    for arg in args.plugin_name:
        parameter = arg.split("==")
        plugin_name = parameter[0]

        plugin_info = plugins.plugin_info(plugin_name)

        if len(parameter) >= 2:
            plugin_version = parameter[1]
            if not plugin_version:
                echo.critical("Missing version")
                exit(1)
        else:
            plugin_version = None

        if plugin_info and not args.force:
            # Plugin already installed
            if plugin_version is None and args.upgrade:
                # Asked for upgrade
                latest = remote.latest(plugin_name, args.pre, args.deprecated)
                if latest and latest.version == plugin_info.version:
                    echo.alert(f"\t{plugin_name} is already at latest version")
                    continue
            elif plugin_version is None and plugin_info.version == plugin_version:
                echo.alert(f"\t{plugin_name}=={plugin_version} already installed")
                continue
            elif plugin_version is None:
                echo.alert(f"\t{plugin_name} already installed")
                continue

        try:
            install_version = remote.install(
                plugin_name=plugin_name,
                version=plugin_version,
                plugin_folder=plugin_info.install_folder if plugin_info else None,
                include_prerelease=args.pre,
                include_deprecated=args.deprecated,
                fix_permissions=args.fix_permissions,
            )

        except PluginNotFoundError:
            similars = remote.check_similar_names(plugin_name)
            name = next(similars, None)
            if name:
                echo.info(f"\n{plugin_name} not found. Plugins with similar name:")
                echo.info(f"\t{name}")
                for name in similars:
                    echo.info(name)
            cli.exit(1)
        except PluginManagerError as err:
            echo.critical(f"ERROR: {plugin_name}: {err}")
            cli.exit(1)
        else:
            echo.success(f"\tOk {plugin_name} {install_version}")
            installed += 1

    if installed > 0:
        install_epilog()


# Remove
@command("remove", help="Remove a plugin by its name")
@argument("plugin_name", help="The plugin to remove")
def remove_plugin(args: Namespace):
    # Local needed only, without QGIS version
    plugins = LocalDirectory(get_plugin_path())
    if not plugins.remove(args.plugin_name):
        cli.exit(1)


# Upgrade
@command("upgrade", help="Upgrade all plugins installed")
@argument(
    "-f",
    "--force",
    action="store_true",
    help="Force reinstall all plugins",
)
@argument(
    "--fix-permissions",
    action="store_true",
    help="Set files permissions to 0644",
)
@argument(
    "--pre",
    action="store_true",
    env="QGIS_PLUGIN_MANAGER_INCLUDE_PRERELEASE",
    help=(
        "Include pre-release, development and experimental versions. By default,\ninstall only stable version"
    ),
)
@argument("--deprecated", action="store_true", help="Include deprecated versions")
def upgrade_plugins(args: Namespace):
    """Upgrade all plugins for which a
    newer version is available
    """
    plugin_path = get_plugin_path()

    qgis = qgis_server_version()
    remote = Remote(plugin_path, qgis_version=qgis)
    plugins = LocalDirectory(plugin_path)
    folders = plugins.plugin_list()

    # Check for ignored plugins
    ignored_plugins = []
    plugin_ignore_file = plugin_path.joinpath("ignorePlugins.list")
    if plugin_ignore_file.exists():
        with open(plugin_ignore_file, encoding="utf8") as f:
            ignored_plugins = [plugin.rstrip() for plugin in f.readlines()]

    installed = 0
    failures = 0

    for folder in folders:
        plugin_info = plugins.plugin_info(folder)
        if not plugin_info:
            echo.debug(f"No plugin found for {folder}")
            continue

        if plugin_info.name in ignored_plugins:
            echo.alert(f"{plugin_info.name:<25}\tIgnored")
            continue

        if not args.force:
            latest = remote.latest(plugin_info.name, args.pre, args.deprecated)
            if latest and latest.version == plugin_info.version:
                echo.success(f"\t\u274e {plugin_info.name:<25} {plugin_info.version!s:<12}\tUnchanged")
                continue
            elif latest is None:
                echo.alert(f"\t\u26a0\ufe0f {plugin_info.name}\tRemoved from repository")
                continue
            version: Optional[str] = str(latest.version) if latest else None
        else:
            version = None

        # Need to check version
        try:
            install_version = remote.install(
                plugin_name=plugin_info.name,
                version=version,
                plugin_folder=plugin_info.install_folder,
                fix_permissions=args.fix_permissions,
                include_prerelease=args.pre,
                include_deprecated=args.deprecated,
            )
        except PluginNotFoundError:
            echo.alert(f"\t\u26a0\ufe0f {plugin_info.name:aa<25}\tNot found")
        except PluginManagerError as err:
            failures += 1
            echo.critical(f"\t\u274c {plugin_info.name:<25}\tError: {err}")
        else:
            installed += 1
            echo.success(f"\t\u2705 {plugin_info.name:<25} {install_version:<12}\tInstalled")

    if failures > 0:
        echo.alert(f"Command terminated with {failures} errors")
    if installed > 0:
        install_epilog()


@command("remotes", help="List all remote sources")
def list_remote_servers(args: Namespace):
    remote = Remote(get_plugin_path(), qgis_server_version())
    remote.print_list()


# Update
@command("update", help="Update all index files")
def update_index(args: Namespace):
    remote = Remote(get_plugin_path(), qgis_server_version())
    if not remote.update():
        cli.exit(1)


# Cache (Deprecated)
@command("cache", help="Look for available plugin is the cache - Deprecated")
@argument("plugin_name", help="The plugin to look for")
def plugin_versions_deprecated(args: Namespace):
    """This command is deprecated, please use the 'versions' command instead."""
    echo.alert("Warning: this command is deprecated in favor of the 'versions' command")
    args.pre = False
    args.deprecated = False
    plugin_versions_impl(args)


# Versions
@command("versions", help="Look for available plugin versions")
@argument("plugin_name", help="The plugin to look for")
@argument(
    "--pre",
    action="store_true",
    env="QGIS_PLUGIN_MANAGER_INCLUDE_PRERELEASE",
    help=(
        "Include pre-release, development and experimental versions. By default,\ndisplay only stable version"
    ),
)
@argument(
    "--format",
    choices=("columns", "list", "json"),
    default="columns",
    help="Select the output format",
)
@argument("--deprecated", action="store_true", help="Include deprecated versions")
def plugin_versions(args: Namespace):
    plugin_versions_impl(args)


def plugin_versions_impl(args: Namespace):
    remote = Remote(get_plugin_path(), qgis_server_version())
    plugins = remote.available_plugins()

    versions = plugins.get(args.plugin_name)
    if versions:

        def results() -> Iterator[Plugin]:
            for plugin in versions:
                if plugin.is_pre() and not args.pre:
                    continue
                if plugin.deprecated and not args.deprecated:
                    continue
                yield plugin

        if args.format == "list":
            for plugin in results():
                echo.echo(f"{plugin.name}=={plugin.version}")
        elif args.format == "json":
            print_json(
                results(),
                (
                    ("name", lambda p: p.name),
                    ("version", lambda p: str(p.version)),
                    ("source", lambda p: p.source),
                    ("createDate", lambda p: p.create_date),
                    ("updateDate", lambda p: p.update_date),
                    ("author", lambda p: p.author_name),
                    ("qgisMinimumVersion", lambda p: str(p.qgis_minimum_version)),
                    ("qgisMaximumVersion", lambda p: str(p.qgis_maximum_version)),
                    ("deprecated", lambda p: p.deprecated),
                    ("experimental", lambda p: p.experimental),
                    ("server", lambda p: p.server),
                    ("tags", lambda p: p.tags),
                    ("trusted", lambda p: p.trusted),
                ),
            )
        else:
            from datetime import datetime

            def display_status(p: Plugin) -> str:
                st: Sequence[str] = ()
                if p.server:
                    st = (*st, "S")
                if p.deprecated:
                    st = (*st, "D")
                if p.experimental:
                    st = (*st, "X")
                if p.trusted:
                    st = (*st, "T")
                return "".join(st)

            def format_date(datestr: Optional[str]) -> str:
                if datestr:
                    dt = datetime.fromisoformat(datestr)
                    return f"{dt:%b %d %Y %H:%M}"
                else:
                    return ""

            echo.success(f"{args.plugin_name}\n")
            print_table(
                tuple(results()),
                (
                    ("Version", lambda p: str(p.version)),
                    ("QGIS min", lambda p: str(p.qgis_minimum_version or "")),
                    ("Status", display_status),
                    ("Source", lambda p: p.source or ""),
                ),
            )

            echo.info("\nStatus: S = Server, X = Experimental, D = Deprecated, T = Trusted")
    else:
        echo.alert(f"No versions found for '{args.plugin_name}'")
        cli.exit(1)


# Search
@command("search", help="Search for plugins")
@argument("plugin_name", help="Search in tags and plugin names")
@argument("--server", action="store_true", help="Consider only server plugins")
@argument("--trusted", action="store_true", help="Consider only trusted plugins")
@argument(
    "--pre",
    action="store_true",
    env="QGIS_PLUGIN_MANAGER_INCLUDE_PRERELEASE",
    help="""
        Include pre-release, development and experimental versions.
        By default, search only stable versions.
    """,
)
@argument("--deprecated", action="store_true", help="Include deprecated versions")
@argument("--latest", action="store_true", help="Consider only latest versions")
def search_plugin(args: Namespace):
    remote = Remote(get_plugin_path(), qgis_server_version())

    def pred(p):
        if args.trusted and not p.trusted:
            return False
        if args.server and not p.server:
            return False
        if not args.pre and p.is_pre():
            return False
        if not args.deprecated and p.deprecated:
            return False
        return True

    found = 0
    for name, version in remote.search(args.plugin_name, predicat=pred, latest=args.latest):
        echo.echo(f"{name}=={version}")
        found += 1
    if not found:
        echo.info("No plugins found")
    else:
        echo.info(f"\nFound {found} plugins")


# Check
@command("check", help="Check compatibility of installed plugins with QGIS version")
@argument("-v", "--version", help="QGIS version to check against")
@argument(
    "--format",
    choices=("columns", "json"),
    default="columns",
    help="Select the output format",
)
def check_qgis_compat(args: Namespace):
    """If version is not specified then check against the current QGIS
    installation
    """

    def get_version() -> Version:
        ver = args.version if args.version else qgis_server_version()
        if not ver:
            exit(1)
        try:
            return get_semver_version(ver)
        except Exception as e:
            echo.critical(f"{e}")
            exit(1)

    version = get_version()

    plugins = LocalDirectory(get_plugin_path())

    def infos():
        for folder, name in sorted(
            plugins.plugin_list().items(),
            key=lambda p: p[1],
        ):
            info = plugins.plugin_info(folder)
            if info:
                yield info

    if args.format == "json":
        print_json(
            infos(),
            (
                ("name", lambda p: p.name),
                ("version", lambda p: str(p.version)),
                ("qgisVersion", lambda _: str(version)),
                ("canUse", lambda p: p.check_qgis_version(version)),
            ),
        )
    else:
        print_table(
            tuple(infos()),
            (
                ("Name", lambda p: p.name),
                ("Version", lambda p: str(p.version)),
                (
                    f"QGIS {version}",
                    lambda p: "Yes" if p.check_qgis_version(version) else "No",
                ),
            ),
        )


def main() -> None:
    """Main function for the CLI menu."""

    args = cli.parse_args()
    if "func" not in args:
        cli.print_help()
        cli.exit(1)
    else:
        echo.set_verbose_mode(args.verbose)
        try:
            args.func(args)
        except PluginManagerError as e:
            echo.critical(f"{e}")


if __name__ == "__main__":
    main()



import base64
import os
import platform
import re
import shutil
import urllib
import urllib.request
import zipfile

from pathlib import Path
from typing import (
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import parse_qs, unquote, urlencode, urlparse, urlunparse
from xml.etree.ElementTree import parse

from semver import Version

from qgis_plugin_manager import echo
from qgis_plugin_manager.definitions import Plugin
from qgis_plugin_manager.utils import (
    PluginManagerError,
    get_semver_version,
    getenv_bool,
    similar_names,
    sources_file,
)

PluginDict = Dict[str, Tuple[Plugin, ...]]


class PluginNotFoundError(PluginManagerError):
    pass


class Remote:
    def __init__(self, folder: Path, qgis_version: Optional[str] = None):
        """Constructor."""
        self.folder = folder
        self.list: List[str] = []
        self.setting_error = False
        self.qgis_version = qgis_version

        self._list_plugins: PluginDict = {}

        self.list_remote()

    def user_agent(self) -> str:
        """User agent."""
        # https://github.com/3liz/qgis-plugin-manager/issues/66
        # https://lists.osgeo.org/pipermail/qgis-user/2024-May/054439.html
        if not self.qgis_version:
            raise PluginManagerError("QGIS version required")
        else:
            qgis_version = self.qgis_version

        qgis_version = qgis_version.replace(".", "")
        return f"Mozilla/5.0 QGIS/{self.qgis_version}/{platform.system()}"

    def check_remote_cache(self) -> bool:
        """Return if the remote is ready to be parsed."""
        if getenv_bool("QGIS_PLUGIN_MANAGER_SKIP_SOURCES_FILE"):
            return True

        source_list = sources_file(self.folder)
        if not source_list.exists():
            if not self.setting_error:
                echo.critical(f"The {source_list.absolute()} file does not exist")
                echo.info("Use the 'init' command to create the file")
                self.setting_error = True
                return False

        cache = self.cache_directory()
        for server in self.list:
            filename = self.server_cache_filename(cache, server)

            if not filename.exists():
                if not self.setting_error:
                    echo.critical(
                        "The 'update' command has not been done before. "
                        f"The repository {server} has not been fetched before."
                    )
                    self.setting_error = True
                    return False

        return True

    def remote_list(self) -> list:
        return self.list

    def list_remote(self) -> list:
        """Return the list of remotes configured.

        The token [VERSION] is replaced by the current version X.YY
        """
        source_list = sources_file(self.folder)
        if not source_list.exists():
            return []

        qgis_version = self.check_qgis_dev_version(self.qgis_version)

        with source_list.open(encoding="utf8") as f:
            for line in f.readlines():
                raw_line = line.strip()
                if not raw_line:
                    # Empty line
                    continue

                if line.startswith("#"):
                    # Commented line
                    continue

                if "[VERSION]" in raw_line:
                    if not qgis_version:
                        echo.alert(
                            f"Skipping line '{raw_line}' because it has a "
                            "token [VERSION] but "
                            "no QGIS version could be detected."
                        )
                        continue

                    raw_line = raw_line.replace("[VERSION]", f"{qgis_version[0]}.{qgis_version[1]}")

                self.list.append(raw_line)

        return self.list

    def cache_directory(self) -> Path:
        """Return the cache directory.

        The default one, or the one defined by environment variable.
        """
        env_path = os.getenv("QGIS_PLUGIN_MANAGER_CACHE_DIR")
        if env_path:
            return Path(env_path)

        return Path(self.folder / ".cache_qgis_plugin_manager")

    def print_list(self):
        """Print in the console the list of remotes."""

        echo.echo("List of remotes :\n")
        if len(self.list):
            echo.echo("\n".join([self.public_remote_name(s) for s in self.list]))
        else:
            echo.alert("No remote configured")

    def latest(
        self,
        name: str,
        include_prerelease: bool = False,
        include_deprecated: bool = False,
        *,
        qgis_version: Optional[Union[str, Version]] = None,
    ) -> Optional[Plugin]:
        if qgis_version and isinstance(qgis_version, str):
            qgis_version = get_semver_version(qgis_version)

        plugin = None
        for plugin in self.available_plugins().get(name, ()):
            if plugin.is_pre() and not include_prerelease:
                continue
            elif plugin.deprecated and not include_deprecated:
                continue
            elif qgis_version is not None and not plugin.check_qgis_version(qgis_version):
                continue
            else:
                break
        else:
            plugin = None
        return plugin

    def update(self) -> bool:
        """For each remote, it updates the XML file."""

        # Clear plugin list
        self._list_plugins = {}

        if not self.list:
            echo.critical("\tNo remote found.")
            return False

        cache = self.cache_directory()
        if cache.exists():
            try:
                shutil.rmtree(cache)
            except OSError as e:
                # https://github.com/3liz/qgis-plugin-manager/issues/53
                echo.critical(f"{e}")
                return False

        cache.mkdir()

        flag = False
        for server in self.list:
            echo.info(f"Downloading {self.public_remote_name(server)}…")
            url, login, password = self.credentials(server)
            headers = {
                "User-Agent": self.user_agent(),
            }
            if login:
                token = base64.b64encode(f"{login}:{password}".encode())
                headers["Authorization"] = f"Basic {token.decode()}"
            request = urllib.request.Request(url, headers=headers)
            try:
                f = urllib.request.urlopen(request)
            except urllib.error.HTTPError as e:
                echo.critical(f"ERROR: {e}")
                continue
            except urllib.error.URLError as e:
                echo.critical(f"ERROR: {e}")
                continue

            filename = self.server_cache_filename(cache, server)

            # Binary mode does not support encoding parameter
            try:
                with open(filename, "wb") as output:
                    output.write(f.read())
            except PermissionError:
                # https://github.com/3liz/qgis-plugin-manager/issues/53
                echo.critical("The directory is not writable.")
                return False

            echo.success("\tOk")
            flag = True

        return flag

    def plugin_collection_files(self) -> Iterator[Tuple[str, Path]]:
        """Returns the list of plugins XML file in the cache folder."""
        cache = self.cache_directory()
        for source in self.list:
            coll = self.server_cache_filename(cache, source)
            if not coll.exists():
                raise PluginManagerError("File cache missing: please run the 'update' command")
            yield source, coll

    def available_plugins(self) -> PluginDict:
        """Populates the list of available plugins, in all XML files."""
        if not self._list_plugins:
            for source, xml_file in self.plugin_collection_files():
                self._parse_xml(xml_file, self._list_plugins, source)
        return self._list_plugins

    def _parse_xml(self, xml_file: Path, plugins: PluginDict, source: Optional[str] = None):
        """Parse the given XML file."""

        tree = parse(xml_file.absolute())
        root = tree.getroot()
        for elem in root:
            plugin = Plugin.from_xml_element(elem, source)

            name = plugin.name
            versions = plugins.get(name)
            if versions:
                # Store as decreasing version order (latest first)
                if plugin.version < versions[-1].version:
                    plugins[name] = (*versions, plugin)
                elif plugin.version > versions[0].version:
                    plugins[name] = (plugin, *versions)
                else:  # Need to sort
                    for i, p in enumerate(versions):
                        if plugin.version > p.version:
                            plugins[name] = (*versions[:i], plugin, *versions[i:])
                            break
                        elif plugin.version == p.version:
                            # SEMVER does not take build into account
                            # but here build are meaningfull because QGIS
                            # plugin version does not stick to SEMVER scheme
                            #
                            # Use lexicographic comparison
                            if (plugin.version.build or "") > (p.version.build or ""):
                                plugins[name] = (*versions[:i], plugin, *versions[i:])
                                break
                    else:
                        plugins[name] = (*versions, plugin)
            else:
                plugins[name] = (plugin,)

    def search(
        self,
        search_string: str,
        strict: bool = True,
        predicat: Optional[Callable[[Plugin], bool]] = None,
        latest: bool = False,
    ) -> Iterator[Tuple[str, Version]]:
        """Search in plugin names and tags."""
        # strict is used in tests to not check if the remote is ready
        if strict and not self.check_remote_cache():
            return

        self.available_plugins()

        results = set()
        for plugin_name, versions in self._list_plugins.items():
            if plugin_name not in results:
                for plugin in versions:
                    if predicat is not None and not predicat(plugin):
                        continue
                    if next(similar_names(search_string, plugin.search), None):
                        results.add(plugin_name)
                        yield plugin.name, plugin.version
                        break
                    if latest:  # Don't look at previous versions
                        break

    def check_similar_names(self, name: str) -> Iterator[str]:
        yield from similar_names(name, self._list_plugins.keys())

    def install(
        self,
        plugin_name: str,
        version: Optional[str] = None,
        include_prerelease: bool = False,
        include_deprecated: bool = False,
        remove_zip: bool = True,
        fix_permissions: bool = False,
        plugin_folder: Optional[str] = None,
    ) -> str:
        """Install the plugin with a specific version.

        Default version is latest.
        """
        self.available_plugins()

        if not self.check_remote_cache():
            raise PluginManagerError("No remote data")

        if not self._list_plugins:
            raise PluginManagerError("No available plugins")

        plugin = None

        # Check for requested version otherwise get the latest
        if version:
            # Find version
            versions = self._list_plugins.get(plugin_name)
            if versions:
                requested_ver = Version.parse(version)
                plugin = next((p for p in versions if p.version == requested_ver), None)
        else:
            plugin = self.latest(
                plugin_name,
                include_prerelease,
                include_deprecated,
            )

        if not plugin:
            echo.alert(f"No matching plugin found for {plugin_name}.")
            raise PluginNotFoundError()

        url = plugin.download_url
        file_name = plugin.file_name

        # Check consistency
        if not file_name:
            raise PluginManagerError(f"Incomplete plugin data: file_name: {plugin})")

        if not url:
            raise PluginManagerError(f"Incomplete plugin data: url: {plugin})")

        echo.debug("Downloading {} from {}", file_name, url)
        zip_file = self._download_zip(url, plugin_name, file_name)

        # Removing existing plugin folder if needed
        if plugin_folder:
            existing = self.folder.joinpath(plugin_folder)
            if existing.exists():
                echo.debug(f"{plugin_name}: Removing existing installation: {existing}")
                try:
                    shutil.rmtree(existing)
                except OSError as e:
                    # https://github.com/3liz/qgis-plugin-manager/issues/53
                    zip_file.unlink()
                    raise PluginManagerError(f"{e}")

        if not zip_file.exists():
            raise PluginManagerError(
                f"The zip file does not exist : {zip_file.absolute()}",
            )
        # Extracting the zip in the folder
        echo.debug(f"Extracting {zip_file.name}")
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(self.folder)

        if remove_zip:
            # Removing the zip file
            zip_file.unlink()

        # Set permissions to 0644 for files, 0755 for directories
        if fix_permissions:
            echo.debug("Fixing files permissions to 0644")
            for p in self.folder.glob("**"):
                if p.is_dir():
                    p.chmod(0o755)
                else:
                    p.chmod(0o644)

        return str(plugin.version)

    def _download_zip(
        self,
        url: str,
        plugin_name: str,
        file_name: str,
    ) -> Path:
        """Download the ZIP"""
        if url.startswith("file:"):
            zip_file = Path(unquote(urlparse(url).path))
        else:
            headers = {
                "User-Agent": self.user_agent(),
            }
            request = urllib.request.Request(url, headers=headers)
            try:
                f = urllib.request.urlopen(request)
            except urllib.error.HTTPError as e:
                if e.code == 401:
                    echo.debug("Authentication required")
                    for _, login, password in self.all_credentials():
                        # Hack to try all logins until we find the one working…
                        token = base64.b64encode(f"{login}:{password}".encode())
                        headers["Authorization"] = f"Basic {token.decode()}"

                        request = urllib.request.Request(url, headers=headers)
                        try:
                            f = urllib.request.urlopen(request)
                            break
                        except urllib.error.HTTPError:
                            continue
                    else:
                        raise PluginManagerError("Failed to download plugin")
                elif e.code == "404":
                    raise PluginManagerError("Plugin not found.")
                else:
                    raise PluginManagerError(f"Error downloading plugin: {e}")

            # Saving the zip from the URL
            zip_file = self.folder.joinpath(file_name)

            try:
                with open(zip_file, "wb") as output:
                    output.write(f.read())
            except PermissionError:
                file_path = self.folder.absolute()
                echo.critical(f"Cannot write to \t{file_path}")
                raise

        return zip_file

    @staticmethod
    def check_qgis_dev_version(qgis_version: Optional[str]) -> Optional[List[str]]:
        """Check if the QGIS current version is odd number."""
        if not qgis_version:
            return None

        qgis_version_info = qgis_version.split(".")
        if int(qgis_version_info[1]) % 2 != 0:
            echo.alert(
                f"A QGIS development version is detected : {qgis_version_info[0]}.{qgis_version_info[1]}."
            )
            qgis_version_info[1] = str(int(qgis_version_info[1]) + 1)
            echo.alert(f"If needed, it will use {qgis_version_info[0]}.{qgis_version_info[1]} instead.")
        return qgis_version_info

    @staticmethod
    def server_cache_filename(cache_folder: Path, server: str) -> Path:
        """Return the path for XML file."""
        filename = "".join(x if x.isalnum() else "-" for x in server)
        filename = re.sub(r"\-+", "-", filename)
        return cache_folder.joinpath(f"{filename}.xml")

    @classmethod
    def credentials(cls, server: str) -> Tuple[str, str, str]:
        """Parse for login and password if needed."""
        u = urlparse(server)
        query = parse_qs(u.query, keep_blank_values=True)
        login = query.get("username", "")
        password = query.get("password", "")

        query.pop("username", None)
        query.pop("password", None)

        u = u._replace(query=urlencode(query, True))
        if login and password:
            return urlunparse(u), login[0], password[0]

        return urlunparse(u), "", ""

    def all_credentials(self) -> Iterator[Tuple[str, str, str]]:
        """Dirty hack to get all credentials for now…"""

        for server in self.list:
            url, login, password = self.credentials(server)
            yield url, login, password

    @classmethod
    def public_remote_name(cls, server: str) -> str:
        """Clean a URL from a password if needed."""
        u = urlparse(server)
        query = parse_qs(u.query, keep_blank_values=True)
        if "password" in query.keys():
            query["password"] = ["******"]
        u = u._replace(query=urlencode(query, True))
        return urlunparse(u)



"""Simple façade to console output"""

import sys

verbose = False


def set_verbose_mode(enable: bool):
    global verbose
    verbose = enable
    if verbose:
        print("=Verbose mode activated", file=sys.stderr)


def echo(s: str):
    print(s)


def info(s: str):
    print(s, file=sys.stderr)


def debug(s: str, *args, **kwargs):
    if verbose:
        print(f"\033[34m{s.format(*args, **kwargs)}\033[0m", file=sys.stderr)


def format_success(s: str) -> str:
    return f"\033[92m{s}\033[0m"


def success(s: str):
    print(format_success(s), file=sys.stderr)


def format_alert(s: str) -> str:
    return f"\033[93m{s}\033[0m"


def alert(s: str):
    print(format_alert(s), file=sys.stderr)


def format_critical(s: str) -> str:
    return f"\033[91m{s}\033[0m"


def critical(s: str):
    print(format_critical(s), file=sys.stderr)



import os

from difflib import SequenceMatcher
from itertools import takewhile
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from semver import Version

from qgis_plugin_manager import echo

DEFAULT_REMOTE_REPOSITORY = "https://plugins.qgis.org"


class PluginManagerError(Exception):
    pass


def get_default_remote_repository() -> str:
    return os.getenv("QGIS_PLUGIN_MANAGER_REMOTE_REPOSITORY", DEFAULT_REMOTE_REPOSITORY)


def restart_qgis_server():
    """Restart QGIS Server tip."""
    restart_file = os.getenv("QGIS_PLUGIN_MANAGER_RESTART_FILE")
    if restart_file:
        Path(restart_file).touch()
    else:
        echo.info(f"{echo.format_alert('Tip')} : Do not forget to restart QGIS Server to reload plugins 😎")
        return


def install_epilog():
    # Get current users
    sudo_user = os.environ.get("SUDO_USER")
    user = current_user()
    # Installation done !
    if sudo_user:
        echo.info(f"\nPlugin(s) Installed with super user '{user}'")
    else:
        echo.info(f"\nPlugin(s) Installed with user '{user}'")

    echo.info(
        "Note: check file permissions and owner according to the user running QGIS Server.",
    )

    restart_qgis_server()


def similar_names(expected: str, available: Iterable[str]) -> Iterator[str]:
    """Returns a list of similar names available."""
    for item in available:
        ratio = SequenceMatcher(None, expected.lower(), item.lower()).ratio()
        if ratio > 0.8:
            yield item


def current_user() -> str:
    """Return the current user if possible."""
    import getpass

    user: Optional[Union[str, int]] = None

    try:
        user = getpass.getuser()
        if user:
            return user
    except KeyError:
        pass

    try:
        user = os.getlogin()
        if user:
            return user
    except OSError:
        pass

    user = os.getegid()
    if user:
        return str(user)

    user = os.environ.get("USER")
    if user:
        return user

    user = os.environ.get("UID")
    if user:
        return user

    return "Unknown"


def qgis_server_version() -> Optional[str]:
    """Try to guess the QGIS Server version.

    On linux distro, qgis python packages are installed at standard location
    in /usr/lib/python3/dist-packages
    """
    try:
        from qgis.core import Qgis

        return Qgis.QGIS_VERSION.split("-")[0]
    except ImportError:
        echo.alert("Cannot check version with PyQGIS, check your QGIS installation or your PYTHONPATH")
        echo.info(f"Current user : {current_user()}")
        echo.info(f"PYTHONPATH={os.getenv('PYTHONPATH')}")
        return None


def sources_file(current_folder: Path) -> Path:
    """Return the default path to the "sources.list" file.

    The path by default or if it's defined with the environment variable.
    """
    env_path = os.getenv("QGIS_PLUGIN_MANAGER_SOURCES_FILE")
    if env_path:
        source_file = Path(env_path)
    else:
        source_file = current_folder.joinpath("sources.list")

    return source_file


def get_semver_version(version_str: str) -> Version:
    """Ensure that we get a SemvVer compatible version

    QGIS does not enforce plugin version to be SemVer
    compatible.

    This represents a best effort to convert version strings
    to compatible SemVer version scheme.

    See https://semver.org/

    Examples:
        2.4.0.1    -> 2.4.0+1
        23.2a      -> 23.0.0+2a
        release    -> 0.0.0+release
        0.6-beta.3 -> 0.6.0-beta.3
    """
    for prefix in ("ver.", "ver", "v.", "v"):
        version_str = version_str.removeprefix(prefix)

    # Check if this is SemVer compatible
    try:
        return Version.parse(version_str)
    except Exception:
        pass

    source_str = version_str

    # Split at hyphen, it may be a prerelease definition
    parts, *pre = version_str.split("-", maxsplit=1)
    pre = f"-{pre[0]}" if pre else ""  # type: ignore [assignment]

    # Split parts of the version string
    parts = parts.split(".", maxsplit=3)  # type: ignore [assignment]

    # Collect at most the three first parts that represent a number up to a
    # non-decimal parts.
    ver = tuple(takewhile(lambda part: part.isdecimal(), parts[:3]))
    # Coalesce remaining parts as a build tag if it does not starts
    # with a dash (like a prerelease tag do)
    rest = ".".join(parts[len(ver) :]) + pre  # type: ignore [operator]
    if rest and not rest.startswith("-"):
        rest = f"+{rest}"

    try_again = 2

    while try_again:
        try_again -= 1

        n = len(ver)
        if n == 3:
            version_str = f"{ver[0]}.{ver[1]}.{ver[2]}{rest}"
        elif n == 2:
            version_str = f"{ver[0]}.{ver[1]}.0{rest}"
        elif n == 1:
            version_str = f"{ver[0]}.0.0{rest}"
        elif n == 0:
            version_str = f"0.0.0{rest}"

        try:
            version = Version.parse(version_str)
            if rest:
                echo.debug(
                    "WARNING: using semver compatible scheme: {} (was {})",
                    version_str,
                    source_str,
                )
            break
        except Exception:
            if not try_again:
                raise

            def replace(c: str) -> str:
                return "-" if c == "_" or not c.isalnum() or not c.isascii() else c

            # Semver error
            # Replace non hyphen/no alphanumeric characters
            rest = "".join(replace(c) for c in rest[1:])
            rest = f"+{rest}"

    return version


def getenv_bool(name: str) -> bool:
    return os.getenv(name, "") in ("t", "true", "y", "yes", "1")


T = TypeVar("T")


def print_table(seq: Sequence[T], columns: Sequence[Tuple[str, Callable[[T], str]]]):
    def colw(col: str, key: Callable[[T], str]) -> int:
        return max(max((len(key(n)) for n in seq), default=0), len(col))

    cols = tuple((col, colw(col, key), key) for col, key in columns)
    echo.echo(" ".join("{:<{}}".format(col, w) for col, w, _ in cols))
    echo.echo(" ".join(f"{'':-<{w}}" for _, w, _ in cols))
    for n in seq:
        echo.echo(" ".join("{:<{}}".format(key(n), w) for _, w, key in cols))


def print_json(seq: Iterable[T], columns: Sequence[Tuple[str, Callable[[T], Any]]]):
    import json

    def records():
        for n in seq:
            yield {col: key(n) for col, key in columns}

    echo.echo(json.dumps(tuple(records()), indent=4))



"""
Metadata about the package to easily retrieve information about it.
See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

__all__ = [
    "__author__",
    "__copyright__",
    "__license__",
    "__maintainer__",
    "__summary__",
    "__title__",
    "__uri__",
]

__author__ = "Etienne Trimaille"
__maintainer__ = "3Liz"
__copyright__ = f"2021 {__author__}, 2025 3liz"
__license__ = "GNU General Public License v3.0"
__summary__ = "A tool for downloading/managing QGIS plugins from CLI."
__title__ = "QGIS Plugin Manager"
__uri__ = "https://github.com/3liz/qgis-plugin-manager"



from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
    Protocol,
    Union,
)

from semver import Version

from .utils import get_semver_version


class Element(Protocol):
    attrib: Dict[str, str]

    def __iter__(self): ...


TRUE_VALUES = ("true", "yes", "1")


class Plugin(NamedTuple):
    """Definition of a plugin in the XML file."""

    name: str
    version: Version
    file_name: Optional[str] = None
    download_url: Optional[str] = None
    description: str = ""
    search: List[str] = []  # noqa RUF012
    qgis_minimum_version: Optional[Version] = None
    qgis_maximum_version: Optional[Version] = None
    homepage: Optional[str] = None
    pre_release: Optional[str] = None
    icon: Optional[str] = None
    author_name: Optional[str] = None
    uploaded_by: Optional[str] = None
    create_date: Optional[str] = None
    update_date: Optional[str] = None
    experimental: bool = False
    deprecated: bool = False
    tracker: Optional[str] = None
    repository: Optional[str] = None
    tags: Optional[str] = None
    server: bool = False
    has_processing: bool = False
    has_wps: bool = False
    trusted: bool = False
    install_folder: Optional[str] = None
    source: Optional[str] = None

    def is_pre(self) -> bool:
        return self.version.prerelease is not None or self.experimental

    def check_qgis_version(self, version: Union[str, Version]) -> bool:
        if self.qgis_minimum_version is not None and self.qgis_minimum_version > version:
            return False
        elif self.qgis_maximum_version is not None and self.qgis_maximum_version < version:
            return False
        else:
            return True

    @staticmethod
    def from_xml_element(elem: Element, source: Optional[str] = None) -> "Plugin":
        data: Dict = {"source": source}
        for element in elem:
            if element.tag in Plugin._fields:
                data[element.tag] = element.text

        experimental = data.get("experimental")
        if experimental:
            data["experimental"] = experimental.lower() in TRUE_VALUES

        deprecated = data.get("deprecated")
        if deprecated:
            data["deprecated"] = deprecated.lower() in TRUE_VALUES

        trusted = data.get("trusted")
        if trusted:
            data["trusted"] = trusted.lower() in TRUE_VALUES

        data["name"] = elem.attrib["name"]
        data["version"] = get_semver_version(elem.attrib["version"])

        def maybe_version(ver: Optional[str]) -> Optional[Version]:
            return get_semver_version(ver) if ver else None

        data["qgis_minimum_version"] = maybe_version(data.get("qgis_minimum_version"))
        data["qgis_maximum_version"] = maybe_version(data.get("qgis_maximum_version"))

        # Add more search fields
        tags = []
        data_tags = data.get("tags")
        if data_tags:
            tags = data_tags.split(",")

        name = data["name"]
        search_text = (
            name.lower(),
            name.lower().replace(" ", ""),
            *tags,
            *name.lower().split(" "),
        )

        # Remove duplicates
        data["search"] = list(dict.fromkeys(search_text))

        return Plugin(**data)
